@@ -1,19 +1,20 @@
-import copy, types, sets
+import copy, types, new
+
+try: 
+    set 
+except NameError: 
+    from sets import Set as set     # Python 2.3 fallback 
 
 from reportlab.lib.units import cm
 from reportlab.lib.pagesizes import A4
 from reportlab.lib.colors import black
 from reportlab.lib.enums import TA_LEFT, TA_CENTER, TA_RIGHT
 
-from utils import calculate_size, get_attr_value
+from utils import calculate_size, get_attr_value, landscape, format_date, memoize,\
+        BAND_WIDTH, BAND_HEIGHT, CROSS_COLS, CROSS_ROWS
 from exceptions import EmptyQueryset, ObjectNotFound, ManyObjectsFound,\
         AttributeNotFound, NotYetImplemented
-
-BAND_WIDTH = 'band-width'
-BAND_HEIGHT = 'band-height'
-
-def landscape(page_size):
-    return page_size[1], page_size[0]
+from cache import DEFAULT_CACHE_STATUS, CACHE_BACKEND, CACHE_FILE_ROOT
 
 class GeraldoObject(object):
     """Base class inherited by all report classes, including band, subreports,
@@ -30,7 +31,7 @@ class GeraldoObject(object):
         if 'name' in kwargs:
             self.name = kwargs.pop('name')
 
-    def destroy(self): # XXX
+    def destroy(self):
         try:
             children = self.get_children()
         except (NotYetImplemented, AttributeError):
@@ -77,7 +78,7 @@ class GeraldoObject(object):
             found.extend(ch_found)
 
         # Cleans using a set
-        found = list(sets.Set(found))
+        found = list(set(found))
 
         # Found nothing
         if not found:
@@ -89,6 +90,37 @@ class GeraldoObject(object):
 
         return many and found or found[0]
 
+    def find_by_type(self, typ):
+        """Find child by informed type (and raises an exception if doesn't
+        find).
+        
+        Attributes:
+            
+            * typ - class type to find
+        """
+        found = []
+
+        # Get object children
+        children = self.get_children()
+
+        for child in children:
+            # Child with the type it is searching for
+            if isinstance(child, typ):
+                found.append(child)
+
+            # Search on child's children
+            try:
+                ch_found = child.find_by_type(typ)
+            except ObjectNotFound:
+                ch_found = []
+
+            found.extend(ch_found)
+
+        # Cleans using a set
+        found = list(set(found))
+
+        return found
+
     def get_children(self):
         """Returns all children elements from this one. This must be overriden
         by inherited class."""
@@ -112,9 +144,11 @@ class GeraldoObject(object):
         raise NotYetImplemented()
 
     def get_object_value(self, obj=None, attribute_name=None, action=None):
-        """Override this method to customize the behaviour of object getting
-        its value."""
-        return self.parent.get_object_value(obj, attribute_name, action)
+        """Override this method to customize the behaviour of object getting its value."""
+        try:
+            return self.parent.get_object_value(obj, attribute_name, action)
+        except AttributeError:
+            raise AttributeNotFound()
 
 class BaseReport(GeraldoObject):
     """Basic Report class, inherited and used to make reports adn subreports"""
@@ -137,6 +171,12 @@ class BaseReport(GeraldoObject):
     default_stroke_color = black
     default_fill_color = black
     borders = None
+    
+    # Events (don't make a method with their names, override 'do_*' instead)
+    before_print = None         # |     before render
+    before_generate = None      # |     after render / before generate
+    after_print = None          # V     after generate
+    on_new_page = None
 
     def __init__(self, queryset=None):
         self.queryset = queryset or self.queryset
@@ -186,7 +226,7 @@ class BaseReport(GeraldoObject):
         if not self.queryset:
             return []
 
-        return [object for object in self.queryset]
+        return list(self.queryset)
 
     def format_date(self, date, expression):
         """Use a date format string method to return formatted datetime.
@@ -195,7 +235,8 @@ class BaseReport(GeraldoObject):
         this (until we find a better and agnosthic solution).
         
         Please don't hack this method up. Just override it on your report class."""
-        return date.strftime(expression)
+
+        return format_date(date, expression)
 
     def get_children(self):
         ret = []
@@ -263,6 +304,55 @@ class BaseReport(GeraldoObject):
         called from their children and on..."""
         raise AttributeNotFound
 
+    # Events methods
+    def do_before_print(self, generator):
+        if self.before_print:
+            self.before_print(self, generator)
+
+    def do_before_generate(self, generator):
+        if self.before_generate:
+            self.before_generate(self, generator)
+
+    def do_after_print(self, generator):
+        if self.after_print:
+            self.after_print(self, generator)
+
+    def do_on_new_page(self, page, page_number, generator):
+        if self.on_new_page:
+            self.on_new_page(self, page, page_number, generator)
+
+    def get_variable_value(self, name, system_fields):
+        """Returns the value for a given variable name"""
+        return system_fields.widget.generator.variables[name]
+
+
+# Useful to find declared report classes without manual registration
+_registered_report_classes = []
+
+class ReportMetaclass(type):
+    """This metaclass registers the declared classes to a local variable."""
+    
+    def __new__(cls, name, bases, attrs):
+        new_class = super(ReportMetaclass, cls).__new__(cls, name, bases, attrs)
+
+        # Defines a registration ID
+        if attrs.get('_registered_id', None) is None:
+            new_class._registered_id = '%s.%s'%(new_class.__module__, name)
+
+        # Appends the new class to list of registered report classes
+        if new_class._registered_id != 'geraldo.base.Report':
+            _registered_report_classes.append(new_class)
+
+        return new_class
+
+@memoize
+def get_report_class_by_registered_id(reg_id):
+    for report_class in _registered_report_classes:
+        if getattr(report_class, '_registered_id', None) == reg_id:
+            return report_class
+
+    return None
+
 class Report(BaseReport):
     """This class must be inherited to be used as a new report.
     
@@ -271,6 +361,8 @@ class Report(BaseReport):
     
     Depends on ReportLab to work properly"""
 
+    __metaclass__ = ReportMetaclass
+
     # Report properties
     title = ''
     author = ''
@@ -289,13 +381,33 @@ class Report(BaseReport):
     # SubReports
     subreports = None
 
+    # Look and feel
+    additional_fonts = None
     default_style = None
 
+    # Caching related attributes
+    cache_status = None
+    cache_backend = None
+    cache_prefix = None
+    cache_file_root = None
+
     def __init__(self, queryset=None):
         super(Report, self).__init__(queryset)
 
+        # Default attributes
         self.subreports = self.subreports and list(self.subreports) or []
         self.default_style = self.default_style or {}
+        self.additional_fonts = self.additional_fonts or {}
+
+        # Caching related attributes
+        if self.cache_status is None:
+            self.cache_status = DEFAULT_CACHE_STATUS
+        if self.cache_backend is None:
+            self.cache_backend = CACHE_BACKEND
+        if self.cache_prefix is None:
+            self.cache_prefix = '-'.join([self.__class__.__module__, self.__class__.__name__])
+        if self.cache_file_root is None:
+            self.cache_file_root = CACHE_FILE_ROOT
 
         # Calls the method that set this as parent if their children
         self.set_parent_on_children()
@@ -310,11 +422,54 @@ class Report(BaseReport):
         if not self.print_if_empty and not self.queryset:
             raise EmptyQueryset("This report doesn't accept empty queryset")
 
+        # TODO: use multiprocessing
         # Initialize generator instance
         generator = generator_class(self, *args, **kwargs)
 
         return generator.execute()
 
+    def generate_under_process_by(self, generator_class, *args, **kwargs):
+        """Uses the power of multiprocessing library to run report generation under
+        a Process and save memory consumming, with better use of multi-core servers.
+        
+        This just will work well if you are generating in a destination file or
+        file-like object (i.e. an HttpResponse on Django).
+        
+        It doesn't returns nothing because Process doesn't."""
+
+        import tempfile, random, os
+        from utils import run_under_process
+
+        # Checks 'filename' argument
+        if 'filename' in kwargs and not isinstance(kwargs['filename'], basestring):
+            # Stores file-like object
+            filelike = kwargs.pop('filename')
+
+            # Make a randomic temporary filename
+            chars = map(chr, range(ord('a'), ord('z')) + range(ord('0'), ord('9')))
+            filename = ''.join([random.choice(chars) for c in range(40)])
+            kwargs['filename'] = os.path.join(tempfile.gettempdir(), filename)
+        else:
+            filelike = None
+
+        @run_under_process
+        def generate_report(report, generator_class, *args, **kwargs):
+            # Generate report into response object
+            report.generate_by(generator_class, *args, **kwargs)
+
+        # Run report generation
+        generate_report(self, generator_class, *args, **kwargs)
+
+        # Loads temp file
+        if filelike:
+            # Reads the temp file
+            fp = file(kwargs['filename'])
+            cont = fp.read()
+            fp.close()
+
+            # Writes temp file content in file-like object
+            filelike.write(cont)
+
     def get_page_rect(self):
         """Calculates a dictionary with page dimensions inside the margins
         and returns. It is used to make page borders."""
@@ -359,15 +514,17 @@ class Report(BaseReport):
 class SubReport(BaseReport):
     """Class to be used for subreport objects. It doesn't need to be inherited.
     
-    'queryset_string' must be a string with path for Python compatible queryset.
+    - 'queryset_string' must be a string with path for Python compatible queryset.
+    - 'get_queryset' is an optional lambda attribute can be used in replacement to
+      queryset_string to make more dynamic querysets
     
     Examples:
     
         * '%(object)s.user_permissions.all()'
         * '%(object)s.groups.all()'
         * 'Message.objects.filter(user=%(object)s)'
-        * 'Message.objects.filter(user__id=%(object)s.id)'
-    """
+        * 'Message.objects.filter(user__id=%(object)s.id)'"""
+
     _queryset_string = None
     _parent_object = None
     _queryset = None
@@ -378,6 +535,8 @@ class SubReport(BaseReport):
 
     visible = True
 
+    get_queryset = None # This must be a lambda function
+
     def __init__(self, **kwargs):
         for k,v in kwargs.items():
             # Validates backward incompatibility for 'detail_band'
@@ -397,12 +556,22 @@ class SubReport(BaseReport):
             self.band_detail.is_detail = True
 
     def queryset(self):
-        if not self._queryset and self.parent_object and self.queryset_string:
-            # Replaces the string representer to a local variable identifier
-            queryset_string = self.queryset_string%{'object': 'parent_object'}
+        if not self._queryset:
+            # Lambda function
+            if self.get_queryset:
+                self._queryset = self.get_queryset(self, self.parent_object)
 
-            # Loads the queryset from string
-            self._queryset = eval(
+            # Queryset string
+            elif self.parent_object and self.queryset_string:
+                # Replaces the string representer to a local variable identifier
+                queryset_string = self.queryset_string%{
+                    'object': 'parent_object',  # TODO: Remove in future
+                    'parent': 'parent_object',
+                    'p': 'parent_object', # Just a short alias
+                    }
+
+                # Loads the queryset from string
+                self._queryset = eval(
                     queryset_string,
                     {'parent_object': self.parent_object},
                     )
@@ -464,14 +633,17 @@ class ReportBand(GeraldoObject):
     height = 1*cm
     width = None # Useful only on detail bands
     visible = True
-    borders = {'top': None, 'right': None, 'bottom': None, 'left': None,
-            'all': None}
+    borders = {'top': None, 'right': None, 'bottom': None, 'left': None, 'all': None}
     elements = None
     child_bands = None
     force_new_page = False
     default_style = None
     auto_expand_height = False
     is_detail = False
+    
+    # Events (don't make a method with their names, override 'do_*' instead)
+    before_print = None
+    after_print = None
 
     def __init__(self, **kwargs):
         for k,v in kwargs.items():
@@ -540,6 +712,15 @@ class ReportBand(GeraldoObject):
                 if isinstance(v, GeraldoObject):
                     v.parent = self
 
+    # Events methods
+    def do_before_print(self, generator):
+        if self.before_print:
+            self.before_print(self, generator)
+
+    def do_after_print(self, generator):
+        if self.after_print:
+            self.after_print(self, generator)
+
 class DetailBand(ReportBand):
     """You should use this class instead of ReportBand in detail bands.
     
@@ -568,6 +749,7 @@ class ReportGroup(GeraldoObject):
     attribute_name = None
     band_header = None
     band_footer = None
+    force_new_page = False
 
     def __init__(self, **kwargs):
         for k,v in kwargs.items():
@@ -613,6 +795,10 @@ class Element(GeraldoObject):
     _width = 0
     _height = 0
     visible = True
+    
+    # Events (don't make a method with their names, override 'do_*' instead)
+    before_print = None
+    after_print = None
 
     # 'width' property
     def _get_width(self):
@@ -648,6 +834,10 @@ class Element(GeraldoObject):
         new._height = self._height
         new.visible = self.visible
 
+        # Copy events
+        new.before_print = self.before_print
+        new.after_print = self.after_print
+
         if hasattr(self, 'name'):
             new.name = self.name
 
@@ -689,3 +879,89 @@ class Element(GeraldoObject):
     def set_parent_on_children(self):
         pass
 
+    # Events methods
+    def do_before_print(self, generator):
+        if self.before_print:
+            self.before_print(self, generator)
+
+    def do_after_print(self, generator):
+        if self.after_print:
+            self.after_print(self, generator)
+
+    _repr_for_cache_attrs = ('left','top','height','width','visible')
+    def repr_for_cache_hash_key(self):
+        return unicode(dict([(attr, getattr(self, attr)) for attr in self._repr_for_cache_attrs]))
+
+class ManyElements(GeraldoObject):
+    """Class that makes the objects creation more dynamic."""
+
+    element_class = None
+    count = None
+    start_left = None
+    start_top = None
+    visible = True
+    element_kwargs = None
+
+    def __init__(self, element_class, count, start_left=None, start_top=None,
+            visible=None, **kwargs):
+
+        self.element_class = element_class
+        self.count = count
+        self.start_left = start_left is not None and start_left or self.start_left
+        self.start_top = start_top is not None and start_top or self.start_top
+        self.visible = visible is not None and visible or self.visible
+
+        # Stores the additinal arguments to use when creating the elements
+        self.element_kwargs = kwargs.copy()
+
+    def get_elements(self, cross_cols=None):
+        """Returns the elements (or create them if they don't exist."""
+
+        from cross_reference import CrossReferenceMatrix
+
+        count = self.count
+
+        # Get cross cols
+        if not cross_cols and isinstance(self.report.queryset, CrossReferenceMatrix):
+            cross_cols = self.report.queryset.cols()
+
+            if count == CROSS_COLS:
+                count = len(cross_cols)
+
+        _elements = []
+
+        # Loop for count of elements to be created
+        next_left = self.start_left
+        next_top = self.start_top
+        for num in range(count):
+            kwargs = self.element_kwargs.copy()
+
+            # Set attributes before creation
+            for k,v in kwargs.items():
+                if v == CROSS_COLS:
+                    try:
+                        kwargs[k] = cross_cols[num]
+                    except IndexError:
+                        kwargs[k] = cross_cols[-1]
+                elif isinstance(v, (list,tuple)) and v:
+                    try:
+                        kwargs[k] = v[num]
+                    except IndexError:
+                        kwargs[k] = v[-1]
+
+            # Create the element
+            el = self.element_class(**kwargs)
+
+            # Set attributes after creation
+            if self.start_left is not None: # Maybe we should support distance here
+                el.left = next_left
+                next_left += el.width
+
+            if self.start_top is not None: # Maybe we should support distance here
+                el.top = next_top
+                next_top += el.height
+
+            _elements.append(el)
+
+        return _elements
+
